"""Datatypes for use with L{ShaderProgram}s.

@author: Stephan Wenger
@date: 2012-02-29
"""

import numpy as _np

import glitter.raw as _gl
from glitter.utils.dtypes import Datatype, bool8, int32, uint32, float32, float64

class ShaderDatatype(object):
    _gltype_db = {}
    _db = {}

    def __init__(self, name, _gltype, dtype=int32, shape=1, texture=False, atomic=True):
        self._name = name
        self._gltype = _gltype
        self._dtype = dtype
        self._shape = (shape,) if not hasattr(shape, "__iter__") and not isinstance(shape, str) else tuple(shape)
        self._texture = texture
        self._atomic = atomic

        Datatype._gltype_db[self._gltype] = self
        Datatype._db[self._dtype, self._shape, self._texture, self._atomic] = self

    @classmethod
    def _from_gl(cls, _gltype):
        return Datatype._gltype_db[_gltype]

    def get_value(self, program, location):
        data = _np.empty(self._shape, int32.as_numpy() if self._dtype.is_boolean() else self._dtype.as_numpy())
        if self._dtype == int32 or self._dtype == bool8:
            _gl.glGetUniformiv(program._id, location, _gl.cast(data.ctypes, _gl.glGetUniformiv.argtypes[-1]))
        elif self._dtype == uint32:
            _gl.glGetUniformuiv(program._id, location, _gl.cast(data.ctypes, _gl.glGetUniformuiv.argtypes[-1]))
        elif self._dtype == float32:
            _gl.glGetUniformfv(program._id, location, _gl.cast(data.ctypes, _gl.glGetUniformfv.argtypes[-1]))
        elif self._dtype == float64:
            _gl.glGetUniformdv(program._id, location, _gl.cast(data.ctypes, _gl.glGetUniformdv.argtypes[-1]))
        else:
            raise TypeError("cannot get uniform variable of type %s" % self)
        return data.astype(self._dtype.as_numpy())

    def set_value(self, program, location, value, count):
        dtype = int32 if self._dtype.is_boolean() else self._dtype
        value = _np.ascontiguousarray(value, dtype.as_numpy())
        if count == 1:
            if value.shape != self._shape and value.shape != (count,) + self._shape:
                raise TypeError("shapes do not match")
        elif self._shape == (1,):
            if value.shape != (count,) and value.shape != (count,) + self._shape:
                raise TypeError("shapes do not match")
        else:
            if value.shape != (count,) + self._shape:
                raise TypeError("shapes do not match")
        if len(self._shape) == 1:
            setter = getattr(_gl, "glUniform%d%sv" % (self._shape[0], dtype.charcode))
            with program:
                setter(location, count, _gl.cast(value.ctypes, setter.argtypes[-1]))
        else:
            if self._shape[0] == self._shape[1]:
                setter = getattr(_gl, "glUniformMatrix%d%sv" % (self._shape[0], dtype.charcode))
            else:
                setter = getattr(_gl, "glUniformMatrix%dx%d%sv" % (self._shape[1], self._shape[0], dtype.charcode))
            with program:
                setter(location, count, _gl.GL_TRUE, _gl.cast(value.ctypes, setter.argtypes[-1]))

    def is_signed(self):
        return self.dtype.is_signed

    def is_unsigned(self):
        return not self.dtype.is_unsigned()

    def is_integer(self):
        return self.dtype.is_integer()

    def is_float(self):
        return self.dtype.is_float()

    def is_boolean(self):
        return self.dtype.is_boolean()

    def is_texture(self):
        return self._texture

    def is_atomic(self):
        return self._atomic

    def __str__(self):
        return self._name

shader_float = ShaderDatatype("float", _gl.GL_FLOAT, dtype=float32)
shader_vec2 = ShaderDatatype("vec2", _gl.GL_FLOAT_VEC2, dtype=float32, shape=2)
shader_vec3 = ShaderDatatype("vec3", _gl.GL_FLOAT_VEC3, dtype=float32, shape=3)
shader_vec4 = ShaderDatatype("vec4", _gl.GL_FLOAT_VEC4, dtype=float32, shape=4)
shader_double = ShaderDatatype("double", _gl.GL_DOUBLE, dtype=float64)
shader_dvec2 = ShaderDatatype("dvec2", _gl.GL_DOUBLE_VEC2, dtype=float64, shape=2)
shader_dvec3 = ShaderDatatype("dvec3", _gl.GL_DOUBLE_VEC3, dtype=float64, shape=3)
shader_dvec4 = ShaderDatatype("dvec4", _gl.GL_DOUBLE_VEC4, dtype=float64, shape=4)
shader_int = ShaderDatatype("int", _gl.GL_INT, dtype=int32)
shader_ivec2 = ShaderDatatype("ivec2", _gl.GL_INT_VEC2, dtype=int32, shape=2)
shader_ivec3 = ShaderDatatype("ivec3", _gl.GL_INT_VEC3, dtype=int32, shape=3)
shader_ivec4 = ShaderDatatype("ivec4", _gl.GL_INT_VEC4, dtype=int32, shape=4)
shader_unsigned_int = ShaderDatatype("unsigned int", _gl.GL_UNSIGNED_INT, dtype=uint32)
shader_uvec2 = ShaderDatatype("uvec2", _gl.GL_UNSIGNED_INT_VEC2, dtype=uint32, shape=2)
shader_uvec3 = ShaderDatatype("uvec3", _gl.GL_UNSIGNED_INT_VEC3, dtype=uint32, shape=3)
shader_uvec4 = ShaderDatatype("uvec4", _gl.GL_UNSIGNED_INT_VEC4, dtype=uint32, shape=4)
shader_bool = ShaderDatatype("bool", _gl.GL_BOOL, dtype=bool8)
shader_bvec2 = ShaderDatatype("bvec2", _gl.GL_BOOL_VEC2, dtype=bool8, shape=2)
shader_bvec3 = ShaderDatatype("bvec3", _gl.GL_BOOL_VEC3, dtype=bool8, shape=3)
shader_bvec4 = ShaderDatatype("bvec4", _gl.GL_BOOL_VEC4, dtype=bool8, shape=4)
shader_mat2 = ShaderDatatype("mat2", _gl.GL_FLOAT_MAT2, dtype=float32, shape=(2, 2))
shader_mat3 = ShaderDatatype("mat3", _gl.GL_FLOAT_MAT3, dtype=float32, shape=(3, 3))
shader_mat4 = ShaderDatatype("mat4", _gl.GL_FLOAT_MAT4, dtype=float32, shape=(4, 4))
shader_dmat2 = ShaderDatatype("dmat2", _gl.GL_DOUBLE_MAT2, dtype=float64, shape=(2, 2))
shader_dmat3 = ShaderDatatype("dmat3", _gl.GL_DOUBLE_MAT3, dtype=float64, shape=(3, 3))
shader_dmat4 = ShaderDatatype("dmat4", _gl.GL_DOUBLE_MAT4, dtype=float64, shape=(4, 4))
shader_sampler1D = ShaderDatatype("sampler1D", _gl.GL_SAMPLER_1D, texture=True)
shader_sampler2D = ShaderDatatype("sampler2D", _gl.GL_SAMPLER_2D, texture=True)
shader_sampler3D = ShaderDatatype("sampler3D", _gl.GL_SAMPLER_3D, texture=True)
shader_samplerCube = ShaderDatatype("samplerCube", _gl.GL_SAMPLER_CUBE, texture=True)
shader_sampler1DShadow = ShaderDatatype("sampler1DShadow", _gl.GL_SAMPLER_1D_SHADOW, texture=True)
shader_sampler2DShadow = ShaderDatatype("sampler2DShadow", _gl.GL_SAMPLER_2D_SHADOW, texture=True)
shader_sampler1DArray = ShaderDatatype("sampler1DArray", _gl.GL_SAMPLER_1D_ARRAY, texture=True)
shader_sampler2DArray = ShaderDatatype("sampler2DArray", _gl.GL_SAMPLER_2D_ARRAY, texture=True)
shader_sampler1DArrayShadow = ShaderDatatype("sampler1DArrayShadow", _gl.GL_SAMPLER_1D_ARRAY_SHADOW, texture=True)
shader_sampler2DArrayShadow = ShaderDatatype("sampler2DArrayShadow", _gl.GL_SAMPLER_2D_ARRAY_SHADOW, texture=True)
shader_sampler2DMS = ShaderDatatype("sampler2DMS", _gl.GL_SAMPLER_2D_MULTISAMPLE, texture=True)
shader_sampler2DMSArray = ShaderDatatype("sampler2DMSArray", _gl.GL_SAMPLER_2D_MULTISAMPLE_ARRAY, texture=True)
shader_samplerCubeShadow = ShaderDatatype("samplerCubeShadow", _gl.GL_SAMPLER_CUBE_SHADOW, texture=True)
shader_samplerBuffer = ShaderDatatype("samplerBuffer", _gl.GL_SAMPLER_BUFFER, texture=True)
shader_sampler2DRect = ShaderDatatype("sampler2DRect", _gl.GL_SAMPLER_2D_RECT, texture=True)
shader_sampler2DRectShadow = ShaderDatatype("sampler2DRectShadow", _gl.GL_SAMPLER_2D_RECT_SHADOW, texture=True)
shader_isampler1D = ShaderDatatype("isampler1D", _gl.GL_INT_SAMPLER_1D, texture=True)
shader_isampler2D = ShaderDatatype("isampler2D", _gl.GL_INT_SAMPLER_2D, texture=True)
shader_isampler3D = ShaderDatatype("isampler3D", _gl.GL_INT_SAMPLER_3D, texture=True)
shader_isamplerCube = ShaderDatatype("isamplerCube", _gl.GL_INT_SAMPLER_CUBE, texture=True)
shader_isampler1DArray = ShaderDatatype("isampler1DArray", _gl.GL_INT_SAMPLER_1D_ARRAY, texture=True)
shader_isampler2DArray = ShaderDatatype("isampler2DArray", _gl.GL_INT_SAMPLER_2D_ARRAY, texture=True)
shader_isampler2DMS = ShaderDatatype("isampler2DMS", _gl.GL_INT_SAMPLER_2D_MULTISAMPLE, texture=True)
shader_isampler2DMSArray = ShaderDatatype("isampler2DMSArray", _gl.GL_INT_SAMPLER_2D_MULTISAMPLE_ARRAY, texture=True)
shader_isamplerBuffer = ShaderDatatype("isamplerBuffer", _gl.GL_INT_SAMPLER_BUFFER, texture=True)
shader_isampler2DRect = ShaderDatatype("isampler2DRect", _gl.GL_INT_SAMPLER_2D_RECT, texture=True)
shader_usampler1D = ShaderDatatype("usampler1D", _gl.GL_UNSIGNED_INT_SAMPLER_1D, texture=True)
shader_usampler2D = ShaderDatatype("usampler2D", _gl.GL_UNSIGNED_INT_SAMPLER_2D, texture=True)
shader_usampler3D = ShaderDatatype("usampler3D", _gl.GL_UNSIGNED_INT_SAMPLER_3D, texture=True)
shader_usamplerCube = ShaderDatatype("usamplerCube", _gl.GL_UNSIGNED_INT_SAMPLER_CUBE, texture=True)
shader_usampler1DArray = ShaderDatatype("usampler1DArray", _gl.GL_UNSIGNED_INT_SAMPLER_1D_ARRAY, texture=True)
shader_usampler2DArray = ShaderDatatype("usampler2DArray", _gl.GL_UNSIGNED_INT_SAMPLER_2D_ARRAY, texture=True)
shader_usampler2DMS = ShaderDatatype("usampler2DMS", _gl.GL_UNSIGNED_INT_SAMPLER_2D_MULTISAMPLE, texture=True)
shader_usampler2DMSArray = ShaderDatatype("usampler2DMSArray", _gl.GL_UNSIGNED_INT_SAMPLER_2D_MULTISAMPLE_ARRAY, texture=True)
shader_usamplerBuffer = ShaderDatatype("usamplerBuffer", _gl.GL_UNSIGNED_INT_SAMPLER_BUFFER, texture=True)
shader_usampler2DRect = ShaderDatatype("usampler2DRect", _gl.GL_UNSIGNED_INT_SAMPLER_2D_RECT, texture=True)
shader_image1D = ShaderDatatype("image1D", _gl.GL_IMAGE_1D, texture=True)
shader_image2D = ShaderDatatype("image2D", _gl.GL_IMAGE_2D, texture=True)
shader_image3D = ShaderDatatype("image3D", _gl.GL_IMAGE_3D, texture=True)
shader_image2DRect = ShaderDatatype("image2DRect", _gl.GL_IMAGE_2D_RECT, texture=True)
shader_imageCube = ShaderDatatype("imageCube", _gl.GL_IMAGE_CUBE, texture=True)
shader_imageBuffer = ShaderDatatype("imageBuffer", _gl.GL_IMAGE_BUFFER, texture=True)
shader_image1DArray = ShaderDatatype("image1DArray", _gl.GL_IMAGE_1D_ARRAY, texture=True)
shader_image2DArray = ShaderDatatype("image2DArray", _gl.GL_IMAGE_2D_ARRAY, texture=True)
shader_image2DMS = ShaderDatatype("image2DMS", _gl.GL_IMAGE_2D_MULTISAMPLE, texture=True)
shader_image2DMSArray = ShaderDatatype("image2DMSArray", _gl.GL_IMAGE_2D_MULTISAMPLE_ARRAY, texture=True)
shader_iimage1D = ShaderDatatype("iimage1D", _gl.GL_INT_IMAGE_1D, texture=True)
shader_iimage2D = ShaderDatatype("iimage2D", _gl.GL_INT_IMAGE_2D, texture=True)
shader_iimage3D = ShaderDatatype("iimage3D", _gl.GL_INT_IMAGE_3D, texture=True)
shader_iimage2DRect = ShaderDatatype("iimage2DRect", _gl.GL_INT_IMAGE_2D_RECT, texture=True)
shader_iimageCube = ShaderDatatype("iimageCube", _gl.GL_INT_IMAGE_CUBE, texture=True)
shader_iimageBuffer = ShaderDatatype("iimageBuffer", _gl.GL_INT_IMAGE_BUFFER, texture=True)
shader_iimage1DArray = ShaderDatatype("iimage1DArray", _gl.GL_INT_IMAGE_1D_ARRAY, texture=True)
shader_iimage2DArray = ShaderDatatype("iimage2DArray", _gl.GL_INT_IMAGE_2D_ARRAY, texture=True)
shader_iimage2DMS = ShaderDatatype("iimage2DMS", _gl.GL_INT_IMAGE_2D_MULTISAMPLE, texture=True)
shader_iimage2DMSArray = ShaderDatatype("iimage2DMSArray", _gl.GL_INT_IMAGE_2D_MULTISAMPLE_ARRAY, texture=True)
shader_uimage1D = ShaderDatatype("uimage1D", _gl.GL_UNSIGNED_INT_IMAGE_1D, texture=True)
shader_uimage2D = ShaderDatatype("uimage2D", _gl.GL_UNSIGNED_INT_IMAGE_2D, texture=True)
shader_uimage3D = ShaderDatatype("uimage3D", _gl.GL_UNSIGNED_INT_IMAGE_3D, texture=True)
shader_uimage2DRect = ShaderDatatype("uimage2DRect", _gl.GL_UNSIGNED_INT_IMAGE_2D_RECT, texture=True)
shader_uimageCube = ShaderDatatype("uimageCube", _gl.GL_UNSIGNED_INT_IMAGE_CUBE, texture=True)
shader_uimageBuffer = ShaderDatatype("uimageBuffer", _gl.GL_UNSIGNED_INT_IMAGE_BUFFER, texture=True)
shader_uimage1DArray = ShaderDatatype("uimage1DArray", _gl.GL_UNSIGNED_INT_IMAGE_1D_ARRAY, texture=True)
shader_uimage2DArray = ShaderDatatype("uimage2DArray", _gl.GL_UNSIGNED_INT_IMAGE_2D_ARRAY, texture=True)
shader_uimage2DMS = ShaderDatatype("uimage2DMS", _gl.GL_UNSIGNED_INT_IMAGE_2D_MULTISAMPLE, texture=True)
shader_uimage2DMSArray = ShaderDatatype("uimage2DMSArray", _gl.GL_UNSIGNED_INT_IMAGE_2D_MULTISAMPLE_ARRAY, texture=True)
shader_atomic_uint = ShaderDatatype("atomic_uint", _gl.GL_UNSIGNED_INT_ATOMIC_COUNTER, dtype=uint32, atomic=True)

__all__ = [
    "ShaderDatatype",
    "shader_float",
    "shader_vec2",
    "shader_vec3",
    "shader_vec4",
    "shader_double",
    "shader_dvec2",
    "shader_dvec3",
    "shader_dvec4",
    "shader_int",
    "shader_ivec2",
    "shader_ivec3",
    "shader_ivec4",
    "shader_unsigned_int",
    "shader_uvec2",
    "shader_uvec3",
    "shader_uvec4",
    "shader_bool",
    "shader_bvec2",
    "shader_bvec3",
    "shader_bvec4",
    "shader_mat2",
    "shader_mat3",
    "shader_mat4",
    "shader_dmat2",
    "shader_dmat3",
    "shader_dmat4",
    "shader_sampler1D",
    "shader_sampler2D",
    "shader_sampler3D",
    "shader_samplerCube",
    "shader_sampler1DShadow",
    "shader_sampler2DShadow",
    "shader_sampler1DArray",
    "shader_sampler2DArray",
    "shader_sampler1DArrayShadow",
    "shader_sampler2DArrayShadow",
    "shader_sampler2DMS",
    "shader_sampler2DMSArray",
    "shader_samplerCubeShadow",
    "shader_samplerBuffer",
    "shader_sampler2DRect",
    "shader_sampler2DRectShadow",
    "shader_isampler1D",
    "shader_isampler2D",
    "shader_isampler3D",
    "shader_isamplerCube",
    "shader_isampler1DArray",
    "shader_isampler2DArray",
    "shader_isampler2DMS",
    "shader_isampler2DMSArray",
    "shader_isamplerBuffer",
    "shader_isampler2DRect",
    "shader_usampler1D",
    "shader_usampler2D",
    "shader_usampler3D",
    "shader_usamplerCube",
    "shader_usampler1DArray",
    "shader_usampler2DArray",
    "shader_usampler2DMS",
    "shader_usampler2DMSArray",
    "shader_usamplerBuffer",
    "shader_usampler2DRect",
    "shader_image1D",
    "shader_image2D",
    "shader_image3D",
    "shader_image2DRect",
    "shader_imageCube",
    "shader_imageBuffer",
    "shader_image1DArray",
    "shader_image2DArray",
    "shader_image2DMS",
    "shader_image2DMSArray",
    "shader_iimage1D",
    "shader_iimage2D",
    "shader_iimage3D",
    "shader_iimage2DRect",
    "shader_iimageCube",
    "shader_iimageBuffer",
    "shader_iimage1DArray",
    "shader_iimage2DArray",
    "shader_iimage2DMS",
    "shader_iimage2DMSArray",
    "shader_uimage1D",
    "shader_uimage2D",
    "shader_uimage3D",
    "shader_uimage2DRect",
    "shader_uimageCube",
    "shader_uimageBuffer",
    "shader_uimage1DArray",
    "shader_uimage2DArray",
    "shader_uimage2DMS",
    "shader_uimage2DMSArray",
    "shader_atomic_uint",
]




"""Base classes for OpenGL objects.

@todo: Rethink what C{__del__} and C{__exit__} methods should do when the interpreter exits (e.g. restoring to a deleted object will not work).

@author: Stephan Wenger
@date: 2012-02-29
"""

from functools import wraps as _wraps

import glitter.raw as _gl

def _get_context():
    """Grab the current context from the L{context} module.

    If no current context exists, a new one is created in a window system
    dependent way.

    @note: The import is wrapped in a function definition to avoid infinite
    recursion on import because the L{context} module imports this module.
    """

    from glitter.contexts import context_manager
    return context_manager.current_context or context_manager.create_default_context()

class GLObject(object):
    """Base class for objects that belong to an OpenGL context.

    @ivar _context: The parent context.
    """

    def __init__(self, context=None):
        """Create a new C{GLObject}.

        @param context: The parent context. Uses the current context if C{context} is C{None}.
        @type context: L{Context}
        """

        self._context = context or _get_context()

    @property
    def context(self):
        return self._context

class ManagedObject(GLObject):
    """Base class for objects that can be created and deleted in OpenGL.

    When a C{ManagedObject} instance is garbage collected, the corresponding
    OpenGL object is deleted as well.
    
    For each C{ManagedObject} subclass, the context keeps a database of existing
    objects.

    @ivar _id: ID generated by OpenGL. The ID is only unique per context.
    """

    _generate_id = NotImplemented
    """Constructor function.
    
    Example: C{glGenShader}

    If the function has an C{argtypes} attribute, it may have zero, one, or two
    parameters. If it has one parameter, L{_type} is passed as the parameter,
    and the return value is used as L{_id}.  If it has two parameters, C{1} is
    passed as the first and a C{GLuint} pointer as the second parameter; the
    number returned in this pointer is used as the L{_id}. All other functions
    are simply called and their return values used as L{_id}.
    """
    
    _delete_id = NotImplemented
    """Destructor function.
    
    Example: C{glDeleteShader}

    If the function has an C{argtypes} attribute and two arguments, C{1} is
    passed as the first and a C{GLuint} pointer to L{_id} as the second
    parameter.  Otherwise, L{_id} is passed as the only parameter.
    """

    _type = NotImplemented
    """An optional parameter for C{_generate_id}.
    
    Example: C{GL_VERTEX_SHADER}
    """

    _db = NotImplemented
    """The name of the corresponding object database in the L{Context}.
    
    Example: C{"shaders"}
    """

    def __init__(self, context=None):
        """Create a new C{ManagedObject} using L{_generate_id}.

        @param context: The parent context.
        @type context: L{Context}
        """

        if any(x is NotImplemented for x in (self._generate_id, self._delete_id)):
            raise TypeError("%s is abstract" % self.__class__.__name__)
        super(ManagedObject, self).__init__(context)
        with self._context:
            if hasattr(self._generate_id, "argtypes") and len(self._generate_id.argtypes) == 1:
                self._id = self._generate_id(self._type)
            elif hasattr(self._generate_id, "argtypes") and len(self._generate_id.argtypes) == 2:
                _id = _gl.GLuint()
                self._generate_id(1, _gl.pointer(_id))
                self._id = _id.value
            else:
                self._id = self._generate_id()
        if self._id == 0:
            raise RuntimeError("could not create %s" % self.__class__.__name__)
        if self._db is not NotImplemented:
            getattr(self._context, self._db)._objects[self._id] = self

    def __del__(self):
        """Delete the OpenGL object using L{_delete_id}.

        Any errors will be ignored because the OpenGL module may already have
        been garbage collected when the interpreter exits.
        """

        try:
            with self._context:
                if hasattr(self._delete_id, "argtypes") and len(self._delete_id.argtypes) == 2:
                    self._delete_id(1, _gl.pointer(_gl.GLuint(self._id)))
                else:
                    self._delete_id(self._id)
            self._id = 0
        except:
            pass

class StateMixin(object):
    """Mixin for objects with properties.

    Calling these objects will generate an appropriate L{State} wrapper for use
    in C{with} statements.
    """

    def __call__(self, **kwargs):
        """Return a L{State} wrapper around C{self}."""
        return State(self, do_enter_exit=True, **kwargs)

class BindableObject(GLObject, StateMixin):
    """Base class for objects that can be bound.

    When the object is bound, it returns the object that was previously bound
    to the same target.

    C{BindableObject} instances can be used in C{with} statements so that binding
    and resetting the previous state happens automatically.

    If subclasses define the methods L{_on_bind} or L{_on_release}, these will be
    called by the binding handler in the context whenever the instance is bound
    or unbound, respectively.

    If subclasses define a property C{_bind_value}, this value will be passed to
    the binding function instead of C{self}.

    Binding of an object to different targets (e.g. buffers that are bound to
    different targets, framebuffers that are bound for reading or drawing, and
    textures that are bound to different texture image units) is not covered by
    the C{BindableObject} class.

    @ivar _stack: A stack for storing previous bindings within C{with}
    statements.
    """

    _binding = NotImplemented
    """Name of the corresponding property in the L{Context}.

    Example: C{"array_buffer_binding"}
    """

    _on_bind = NotImplemented
    """Function to call before binding.
    
    L{ShaderProgram}s, for example, bind textures here.
    """
    
    _on_release = NotImplemented
    """Function to call after releasing.

    L{ShaderProgram}s, for example, release textures here.
    """

    _bind_value = NotImplemented
    """Value to pass to the C{_binding} property.

    If C{NotImplemented}, C{self} will be used.
    """

    def __init__(self, context=None):
        """Create a new C{BindableObject}.

        @param context: The parent context.
        @type context: L{Context}
        """

        if any(x is NotImplemented for x in (self._binding,)):
            raise TypeError("%s is abstract" % self.__class__.__name__)
        super(BindableObject, self).__init__(context)
        self._stack = []
    
    def bind(self):
        """Bind the object and return the previously bound object.

        Binding is executed by setting the parent L{Context}'s property named
        in L{_binding}.

        @return: The previous value of the property.
        """

        try:
            old_binding = getattr(self._context, self._binding)
        except AttributeError:
            old_binding = None
        setattr(self._context, self._binding, self if self._bind_value is NotImplemented else self._bind_value)
        return old_binding

    def __enter__(self):
        """Called when a C{with} statement is entered.

        Activates the parent L{Context}, calls L{bind} and stores the returned
        value on the L{_stack}.
        """

        self._context.__enter__()
        self._stack.append(self.bind())
        return self

    def __exit__(self, type, value, traceback):
        """Called when a C{with} statement is exited.

        Restores the previous binding from the L{_stack} by setting the
        attribute named in L{_binding} in the parent L{Context} and deactivates
        the parent L{Context}.
        """

        setattr(self._context, self._binding, self._stack.pop())
        self._context.__exit__(type, value, traceback)

class State(GLObject, StateMixin):
    """Context manager to add binding semantics to context property changes.

    To set properties of the context and automatically reset it to its old
    value later, use a C{with} statement with a C{State} object.
    """

    def __init__(self, context=None, do_enter_exit=False, **kwargs):
        """Create a C{State} object for use in C{with} statements.

        When entering the C{with} statement, the properties of C{context} given
        in C{kwargs} will be set to their respective values. On exiting the
        C{with} statement, the old value will be restored.

        If C{context} defines C{__enter__} and C{__exit__} methods, this will
        be called on entering and exiting the C{with} block, respectively,
        if C{do_enter_exit} is C{True}.

        No guarantee is made about the order in which the properties are set,
        but they are guaranteed to be reset in reverse order.

        For convenience, C{__getattr__} and C{__setattr__} calls are redirected
        to C{context}, so that C{with State(...) as x:} works as expected.

        @param context: The context object on which to set the properties, or the current context if it is C{None}.
        @type context: L{Context} or any other object with attributes
        @param do_enter_exit: Whether to call C{context}'s C{__enter__} and C{__exit__} methods.
        @type do_enter_exit: C{bool}
        @param kwargs: A dictionary of property names and their values.
        @type kwargs: C{dict}
        """

        super(State, self).__init__(context)
        self._do_enter_exit = do_enter_exit
        self._properties = kwargs
        self._stack = []

    def __enter__(self):
        if self._do_enter_exit and hasattr(self._context, "__enter__"):
            self._context.__enter__()
        for key, value in list(self._properties.items()):
            try:
                self._stack.append(getattr(self._context, key))
            except AttributeError:
                self._stack.append(None)
            setattr(self._context, key, value)
        return self

    def __exit__(self, type, value, traceback):
        for key in reversed(list(self._properties.keys())):
            setattr(self._context, key, self._stack.pop())
        if self._do_enter_exit and hasattr(self._context, "__exit__"):
            self._context.__exit__(type, value, traceback)

    def __getattr__(self, key):
        if key == "_stack" or hasattr(self, "_stack"): # constructor not done yet
            return super(State, self).__getattr__(key)
        return getattr(self._context, key)

    def __setattr__(self, key, value):
        if not hasattr(self, "_stack"): # constructor not done yet
            return super(State, self).__setattr__(key, value)
        setattr(self._context, key, value)

class BindReleaseObject(GLObject, StateMixin):
    """Base class for objects that can be bound and released.

    C{BindReleaseObject} should be used instead of L{BindableObject} when binding
    and releasing are not performed by setting a property on a context, but by
    custom L{bind} and L{release} methods.

    Subclasses are responsible for restoring any objects previously bound to
    the same target.
    """

    bind = NotImplemented
    """Method for binding C{self} to L{_context}.

    C{bind} takes no arguments.
    """

    release = NotImplemented
    """Method for releasing C{self} from L{_context}.

    C{release} is responsible for restoring any previous binding.

    C{release} takes no arguments.
    """

    def __init__(self, context=None):
        """Create a new C{BindReleaseObject}.

        @param context: The parent context.
        @type context: L{Context}
        """

        super(BindReleaseObject, self).__init__(context)
        if any(x is NotImplemented for x in (self.bind, self.release)):
            raise TypeError("%s is abstract" % self.__class__.__name__)
    
    def __enter__(self):
        """Called when a C{with} statement is entered.

        Activates the parent L{Context} and calls L{bind}.
        """

        self._context.__enter__()
        self.bind()
        return self

    def __exit__(self, type, value, traceback):
        """Called when a C{with} statement is exited.

        Calls L{release} and deactivates the parent L{Context}.
        """

        self.release()
        self._context.__exit__(type, value, traceback)

def with_obj(obj, f):
    """Create a wrapper that executes C{f} in a C{with obj:} block."""
    @_wraps(f)
    def wrapper(*args, **kwargs):
        with obj:
            return f(*args, **kwargs)
    return wrapper

__all__ = ["GLObject", "ManagedObject", "BindableObject", "BindReleaseObject", "State", "StateMixin", "with_obj"]

